# -*- coding: utf-8 -*-

"""Implementations of the identify algorithm from Shpitser and Pearl."""

from typing import List, Set, Tuple, Union

from ananke.graphs import ADMG
from ananke.identification import OneLineID

from .dsl import (
    CounterfactualVariable,
    Distribution,
    Intervention,
    Probability,
    Variable,
)
from .graph import NxMixedGraph

__all__ = [
    "is_identifiable",
]


def _get_treatments(variables: Set[Variable]) -> List[str]:
    return list(
        {variable.name for variable in variables if isinstance(variable, Intervention)}
    )


def _get_outcomes(variables: Set[Variable]) -> List[str]:
    return list(
        {
            variable.name
            for variable in variables
            if not isinstance(variable, Intervention)
        }
    )


def _all_counterfactual(distribution: Distribution) -> bool:
    return all(
        isinstance(variable, CounterfactualVariable)
        for variable in distribution.children
    )


def _all_intervened_same(distribution: Distribution) -> bool:
    return 1 == len(
        {
            variable.interventions  # type:ignore
            for variable in distribution.children
        }
    )


def _get_to(query: Distribution) -> Tuple[List[str], List[str]]:
    if not _all_counterfactual(query):
        raise ValueError(
            "all variables in input distribution should be counterfactuals"
        )

    if not _all_intervened_same(query):
        raise ValueError("not all variables are invervened on the same")

    treatments = [
        intervention.name
        for intervention in query.children[0].interventions  # type:ignore
    ]
    outcomes = [variable.name for variable in query.children]
    return treatments, outcomes


def is_identifiable(
    graph: Union[ADMG, NxMixedGraph], query: Union[Probability, Distribution]
) -> bool:
    """Check if the expression is identifiable.

    :param graph: Either an Ananke graph or y0 NxMixedGraph that can be converted to an Ananke graph
    :param query: A probability distribution with the following properties:

        1. There are no conditions
        2. All variables are counterfactuals
        3. The set of interventions on each counterfactual variable are the same
    :raises ValueError: If the given probability distribution does not meet the above properties
    :returns: Whether the graph is identifiable under the causal query.

    Example non-identifiable scenario:

    .. code-block:: python

        from y0.graph import NxMixedGraph
        from y0.dsl import P, X, Y

        graph = NxMixedGraph()
        graph.add_directed_edge('X', 'Z')
        graph.add_directed_edge('Z', 'Y')
        graph.add_undirected_edge('X', 'Z')

        assert not is_identifiable(graph, P(Y @ ~X))

    Example identifiable scenario:

    .. code-block:: python

        from y0.graph import NxMixedGraph
        from y0.dsl import P, X, Y

        graph = NxMixedGraph()
        graph.add_directed_edge('X', 'Y')
        graph.add_directed_edge('X', 'Z')
        graph.add_directed_edge('Z', 'Y')
        graph.add_undirected_edge('Y', 'Z')

        assert is_identifiable(graph, P(Y @ ~X))
    """
    if isinstance(graph, NxMixedGraph):
        graph = graph.to_admg()

    if isinstance(query, Probability):
        query = query.distribution

    if query.is_conditioned():
        raise ValueError("input distribution should not have any conditions")

    treatments, outcomes = _get_to(query)

    one_line_id = OneLineID(
        graph=graph,
        treatments=treatments,
        outcomes=outcomes,
    )
    return one_line_id.id()



# -*- coding: utf-8 -*-

"""A parser for Craig-like probability expressions based on :mod:`pyparsing`."""

import logging

from pyparsing import (
    Forward,
    Group,
    OneOrMore,
    Optional,
    ParseException,
    ParseResults,
    StringEnd,
    StringStart,
    Suppress,
)

from .utils import probability_pe, qfactor_pe, variables_pe
from ...dsl import Expression, Fraction, Product, Sum

__all__ = [
    "parse_causaleffect",
    "grammar",
]

logger = logging.getLogger(__name__)
expr = Forward()


def _make_sum(_s, _l, tokens: ParseResults) -> Sum:
    return Sum(
        ranges=tuple(tokens["ranges"].asList()) if "ranges" in tokens else tuple(),
        expression=tokens["expression"],
    )


def _make_frac(_s, _l, tokens: ParseResults) -> Fraction:
    return Fraction(
        numerator=tokens["numerator"],
        denominator=tokens["denominator"],
    )


def _make_product(_s, _l, tokens: ParseResults) -> Expression:
    tokens = tokens.asList()
    if len(tokens) == 1:
        return tokens[0]
    else:
        return Product(tuple(tokens))


# auto-product
rr = OneOrMore(probability_pe | qfactor_pe | expr).setParseAction(_make_product)

sum_pe = (
    Suppress("\\sum_{")
    + Optional(Group(variables_pe).setResultsName("ranges"))
    + Suppress("}")
    + rr.setResultsName("expression")
)
sum_pe.setName("sum")
sum_pe.setParseAction(_make_sum)

fraction_pe = (
    Suppress("\\frac_{")
    + rr.setResultsName("numerator")
    + Suppress("}{")
    + rr.setResultsName("denominator")
    + Suppress("}")
)
fraction_pe.setName("fraction")
fraction_pe.setParseAction(_make_frac)

expr << (probability_pe | qfactor_pe | sum_pe | fraction_pe)

# TODO enable products?

grammar = StringStart() + expr + StringEnd()
grammar.setName("probabilityGrammar")


def parse_causaleffect(s: str) -> Expression:
    """Parse a causaleffect probability expression."""
    try:
        x = grammar.parseString(s)
    except ParseException:
        logger.warning("could not parse %s", s)
        raise
    else:
        return x.asList()[0]



# -*- coding: utf-8 -*-

"""An implementation of Sara Taheri's algorithm for using causal queries for experimental design.

.. seealso:: https://docs.google.com/presentation/d/1klBOjGtRkXOMSDgOCLChBTBJZ0dFxJPn9IPRLAlv_N8/edit?usp=sharing
"""

import itertools as itt
import logging
import textwrap
from pathlib import Path
from typing import Collection, Iterable, List, NamedTuple, Optional, Set, Tuple, Union

import click
import networkx as nx
from ananke.graphs import ADMG
from more_click import verbose_option
from tabulate import tabulate
from tqdm import tqdm

from y0.algorithm.simplify_latent import simplify_latent_dag
from y0.dsl import Expression, P, Variable
from y0.graph import (
    DEFAULT_TAG,
    NxMixedGraph,
    admg_from_latent_variable_dag,
    admg_to_latent_variable_dag,
)
from y0.identify import is_identifiable
from y0.util.combinatorics import powerset

__all__ = [
    "taheri_design_admg",
    "taheri_design_dag",
    "Result",
    "draw_results",
]

logger = logging.getLogger(__name__)


class Result(NamedTuple):
    """Results from the LV-DAG check."""

    identifiable: bool
    identifiability_expr: Optional[Expression]
    pre_nodes: int
    pre_edges: int
    post_nodes: int
    post_edges: int
    latents: List[str]
    observed: List[str]
    lvdag: nx.DiGraph
    admg: ADMG


def taheri_design_admg(
    graph: Union[ADMG, NxMixedGraph],
    cause: str,
    effect: str,
    *,
    tag: Optional[str] = None,
    stop: Optional[int] = None,
) -> List[Result]:
    """Run the brute force implementation of the Taheri Design algorithm on an ADMG.

    :param graph: An ADMG
    :param cause: The node that gets perturbed.
    :param effect: The node that we're interested in.
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :param stop: Largest combination to get (None means length of the list and is the default)
    :return: A list of LV-DAG identifiability results. Will be length 2^(|V| - 2 - # bidirected edges)
    """
    if tag is None:
        tag = DEFAULT_TAG
    if isinstance(graph, NxMixedGraph):
        graph = graph.to_admg()
    dag = admg_to_latent_variable_dag(graph, tag=tag)
    fixed_latent = {node for node, data in dag.nodes(data=True) if data[tag]}
    return _help(
        graph=dag,
        cause=cause,
        effect=effect,
        fixed_observed={cause, effect},
        fixed_latent=fixed_latent,
        tag=tag,
        stop=stop,
    )


def taheri_design_dag(
    graph: nx.DiGraph,
    cause: str,
    effect: str,
    *,
    tag: Optional[str] = None,
    stop: Optional[int] = None,
) -> List[Result]:
    """Run the brute force implementation of the Taheri Design algorithm on a DAG.

    Identify all latent variable configurations inducible over the given DAG that result
    in an identifiable ADMG under the causal query corresponding to the given cause/effect.

    :param graph: A regular DAG
    :param cause: The node that gets perturbed.
    :param effect: The node that we're interested in.
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :param stop: Largest combination to get (None means length of the list and is the default)
    :return: A list of LV-DAG identifiability results. Will be length 2^(|V| - 2)
    """
    return _help(
        graph=graph,
        cause=cause,
        effect=effect,
        fixed_observed={cause, effect},
        tag=tag,
        stop=stop,
    )


def _help(
    graph: nx.DiGraph,
    cause: str,
    effect: str,
    *,
    fixed_observed: Optional[Collection[str]] = None,
    fixed_latent: Optional[Collection[str]] = None,
    tag: Optional[str] = None,
    stop: Optional[int] = None,
) -> List[Result]:
    return [
        _get_result(
            lvdag=lvdag,
            latents=latents,
            observed=observed,
            cause=cause,
            effect=effect,
            tag=tag,
        )
        for latents, observed, lvdag in iterate_lvdags(
            graph,
            fixed_observed=fixed_observed,
            fixed_latents=fixed_latent,
            tag=tag,
            stop=stop,
        )
    ]


def _get_result(
    lvdag,
    latents,
    observed,
    cause,
    effect,
    *,
    tag: Optional[str] = None,
) -> Result:
    # Book keeping
    pre_nodes, pre_edges = lvdag.number_of_nodes(), lvdag.number_of_edges()

    # Apply the robin evans algorithms
    simplify_latent_dag(lvdag, tag=tag)
    post_nodes, post_edges = lvdag.number_of_nodes(), lvdag.number_of_edges()

    # Convert the latent variable DAG to an ADMG
    admg = admg_from_latent_variable_dag(lvdag, tag=tag)

    if cause not in admg.vertices:
        raise KeyError(f"ADMG missing cause: {cause}")
    if effect not in admg.vertices:
        raise KeyError(f"ADMG missing effect: {effect}")

    # Check if the ADMG is identifiable under the (simple) causal query
    query = P(Variable(effect) @ ~Variable(cause))
    identifiable = is_identifiable(admg, query)
    identifiability_expr = None

    return Result(
        identifiable,
        identifiability_expr=identifiability_expr,
        pre_nodes=pre_nodes,
        pre_edges=pre_edges,
        post_nodes=post_nodes,
        post_edges=post_edges,
        latents=sorted(latents),
        observed=sorted(observed),
        lvdag=lvdag,
        admg=admg,
    )


def iterate_lvdags(
    graph: nx.DiGraph,
    fixed_observed: Optional[Collection[str]] = None,
    fixed_latents: Optional[Collection[str]] = None,
    *,
    tag: Optional[str] = None,
    stop: Optional[int] = None,
) -> Iterable[Tuple[Set[str], Set[str], nx.DiGraph]]:
    """Iterate over all possible latent variable configurations for the given graph.

    :param graph: A regularDAG
    :param fixed_observed: Nodes to skip in the power set of all possible latent variables. Often, the cause and
        effect from a causal query will be used here to avoid setting them as latent (since they can not be).
    :param fixed_latents: Nodes to skip in the power set of all possible latent variables. Often, latent nodes
        from ADMG->LV-DAG conversion will go here.
    :param tag: The key for node data describing whether it is latent.
        If None, defaults to :data:`y0.graph.DEFAULT_TAG`.
    :param stop: Largest combination to get (None means length of the list and is the default)
    :yields: latent variable DAGs for all possible latent variable configurations over the original DAG
    """
    if tag is None:
        tag = DEFAULT_TAG

    fixed_observed = set() if not fixed_observed else set(fixed_observed)
    fixed_latents = set() if not fixed_latents else set(fixed_latents)

    inducible_nodes = set(graph)
    inducible_nodes.difference_update(fixed_observed)
    inducible_nodes.difference_update(fixed_latents)

    if stop is None:
        stop = len(inducible_nodes) - 1
    it = powerset(
        sorted(inducible_nodes),
        stop=stop,
        reverse=True,
        use_tqdm=True,
        tqdm_kwargs=dict(desc="LV powerset"),
    )

    graph = graph.copy()
    for node in fixed_observed:
        graph.nodes[node][tag] = False
    for node in fixed_latents:
        graph.nodes[node][tag] = True
    for induced_latents in map(set, it):
        yv = graph.copy()
        for node in inducible_nodes:
            yv.nodes[node][tag] = node in induced_latents
        yield induced_latents, inducible_nodes - induced_latents, yv  # type:ignore


def draw_results(
    results: Iterable[Result],
    path: Union[str, Path, Iterable[str], Iterable[Path]],
    ncols: int = 10,
    x_ratio: float = 4.2,
    y_ratio: float = 4.2,
    max_size: Optional[int] = None,
) -> None:
    """Draw identifiable ADMGs to a file."""
    import matplotlib.pyplot as plt

    if isinstance(path, str):
        path = [path]

    rendered_results = [result for result in results if result.identifiable]
    if max_size is not None:
        rendered_results = [
            result
            for result in results
            if len(result.admg.vertices) - len(result.latents) < max_size
        ]

    logger.debug("rendering %s identifiable queries", rendered_results)

    nrows = 1 + len(rendered_results) // ncols
    fig, axes = plt.subplots(
        ncols=ncols, nrows=nrows, figsize=(ncols * x_ratio, nrows * y_ratio)
    )
    it = itt.zip_longest(axes.ravel(), tqdm(rendered_results, desc="generating chart"))
    for i, (ax, result) in enumerate(it, start=1):
        if result is None:
            ax.axis("off")
        else:
            mixed_graph = NxMixedGraph.from_admg(result.admg)
            title = f"{i}) Latent: " + ", ".join(result.latents)
            if result.identifiability_expr is not None:
                title += f"\n${result.identifiability_expr.to_latex()}$"
            mixed_graph.draw(ax=ax, title="\n".join(textwrap.wrap(title, width=45)))

    fig.tight_layout()

    for _path in tqdm(path, desc="saving"):
        logger.info("saving to %s", _path)
        fig.savefig(_path, dpi=400)


def print_results(results: List[Result], file=None) -> None:
    """Print a set of results."""
    rows = [
        (
            i,
            result.identifiable,
            result.post_nodes - result.pre_nodes,
            result.post_edges - result.pre_edges,
            len(result.latents),
            ", ".join(result.latents),
        )
        for i, result in enumerate(results, start=1)
    ]
    print(
        tabulate(
            rows, headers=["Row", "ID?", "Node Simp.", "Edge Simp.", "N", "Latents"]
        ),
        file=file,
    )


@click.command()
@verbose_option
def main():
    """Run the algorithm on the IGF graph with the PI3K/Erk example."""
    import pystow
    from y0.resources import VIRAL_PATHOGENESIS_PATH
    from y0.graph import NxMixedGraph

    viral_pathogenesis_admg = NxMixedGraph.from_causalfusion_path(
        VIRAL_PATHOGENESIS_PATH
    )

    results = taheri_design_admg(
        viral_pathogenesis_admg, cause="EGFR", effect="CytokineStorm", stop=5
    )
    draw_results(
        results,
        [
            pystow.join("y0", "viral_pathogenesis_egfr.png"),
            pystow.join("y0", "viral_pathogenesis_egfr.svg"),
        ],
    )

    results = taheri_design_admg(
        viral_pathogenesis_admg, cause=r"sIL6R\alpha", effect="CytokineStorm", stop=5
    )
    draw_results(
        results,
        [
            pystow.join("y0", "viral_pathogenesis_sIL6ra.png"),
            pystow.join("y0", "viral_pathogenesis_sIL6ra.svg"),
        ],
    )

    from y0.examples import igf_graph

    results = taheri_design_dag(igf_graph, cause="PI3K", effect="Erk")
    print_results(results)
    draw_results(
        results,
        [
            pystow.join("y0", "ifg_identifiable_configs.png"),
            pystow.join("y0", "ifg_identifiable_configs.svg"),
        ],
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Test falsification of testable implications given a graph."""

import unittest

from y0.algorithm.conditional_independencies import get_conditional_independencies
from y0.algorithm.falsification import falsifications
from y0.examples import asia_example


class TestFalsification(unittest.TestCase):
    """Test the falsifiable implications."""

    def test_asia_from_graph(self):
        """Test the asia graph against data generated from it."""
        graph = asia_example.graph.to_admg()
        df = asia_example.data
        issues = falsifications(graph, df)
        self.assertEqual(0, len(issues))
        self.assertGreater(len(issues.evidence), 0)

    def test_asia_from_list(self):
        """Test the asia graph against data generated from it, passing in the implications to test."""
        graph = asia_example.graph.to_admg()
        df = asia_example.data
        implications = get_conditional_independencies(graph)
        issues = falsifications(implications, df)
        self.assertEqual(0, len(issues))
        self.assertEqual(len(issues.evidence), len(implications))



"""Test the internal y0 DSL parser."""

import unittest

from y0.dsl import A, B, Expression, P, Q, Sum
from y0.parser import parse_y0


class TestInternalParser(unittest.TestCase):
    """Test the internal y0 DSL parser."""

    def test_parse(self):
        """Test the parse function."""
        for expected, s in [
            (
                Sum[B](P(A | B) * P(B)) / Sum(P(A | B) * P(B)),
                "Sum[B](P(A | B) * P(B)) / Sum(P(A | B) * P(B))",
            ),
            (
                Q[A](B) * P(A | B),
                "Q[A](B) * P(A | B)",
            ),
        ]:
            with self.subTest(s=s):
                actual = parse_y0(s)
                self.assertIsInstance(actual, Expression)
                self.assertEqual(
                    expected,
                    actual,
                    msg=f"\nExpected: {expected}\nActual:   {actual}",
                )



# -*- coding: utf-8 -*-

"""Tests for expression predicates."""

import unittest

from y0.dsl import A, B, C, D, P, Sum, Variable, X
from y0.predicates import has_markov_postcondition


class TestMarkovCondition(unittest.TestCase):
    """Tests for checking the markov condition."""

    def test_markov_raises(self):
        """Test the type error is raised on invalid input."""
        for value in [
            Variable("whatever"),
            A @ B,
            A @ ~B,
            "something else",
        ]:
            with self.subTest(value=value), self.assertRaises(TypeError):
                has_markov_postcondition(value)

    def test_markov_postcondition(self):
        """Test the expressions have the markov postcondition."""
        for expression in [
            P(A),
            P(A | B),
            P(A | (B, C)),
            P(A) * P(B),
            P(A) * P(A | B),
            P(A | B) * P(A | C),
            P(A) / P(B),
            P(A) / P(A | B),
            Sum[X](P(A)),
            Sum[X](P(A | B)),
            Sum[X](P(A | B) * P(B)),
        ]:
            with self.subTest(e=expression):
                self.assertTrue(has_markov_postcondition(expression))

    def test_missing_markov_postcondition(self):
        """Test the expressions do not have the markov postcondition."""
        for expression in [
            P(A, B),
            P(A & B | C),
            P(A & D | (B, C)),
            P(A, C) * P(B),
            P(A, C) * P(A | B),
            P(A) * P(A & C | B),
            P(A & C | B) * P(A | C),
            P(A) / P(B & C),
            P(A & C) / P(B),
            P(A) / P(A & C | B),
            Sum[X](P(A, C)),
            Sum[X](P(A & C | B)),
            Sum[X](P(A & C | B) * P(B)),
            Sum[X](P(A | B) * P(B, C)),
        ]:
            with self.subTest(e=str(expression)):
                self.assertFalse(has_markov_postcondition(expression))



# -*- coding: utf-8 -*-

"""Functions that mutate probability expressions."""

from .canonicalize_expr import canonicalize
from .chain import bayes_expand, chain_expand, fraction_expand

__all__ = [
    "canonicalize",
    "chain_expand",
    "fraction_expand",
    "bayes_expand",
]
